﻿// SpringSim class
//
// A class for providing spring motion to movie clips.
//
// If you're curious about the physics, check out the function 'doSimulate'
// below, it is heavily commented.  Everything else here is meant to support
// that function.
//
// Todo: Impart mouse-velocity onto non-fixed dots when dragging.
//        Modify connected spring resting lengths when dragging dots in edit mode.

kDegToRad = Math.PI/180;


function Polygon(pts, clr, curvy, sprs)
{
	this.pts = pts;
	this.clr = clr;
	this.sprs = sprs;
	this.curvy = curvy;
}

function Spring(dot1, dot2, tension)
{
	if (tension == undefined)
		tension = 1;
	this.dot1 = dot1;
	this.dot2 = dot2;
	dot1.vx = dot1.vy = 0;
	dot2.vx = dot2.vy = 0;
	this.dist = this.getLength();
	this.restLen = this.dist*tension;
	// trace(" Spring from " + dot1._name + " --> " + dot2._name + ", " + tension + ", " + this.restLen);
}

Spring.prototype.getLength = function()
{
	var dx = this.dot1._x - this.dot2._x;
	var dy = this.dot1._y - this.dot2._y;
	return Math.sqrt(dx*dx+dy*dy);
}

Spring.prototype.setMuscle = function(amp,phase)
{
	this.isMuscle = true;
	this.amp = amp;
	this.phase = phase;
}

function SpringSim(parent, gravK, stiffK, frictionK, isDraggable)
{
	if (gravK == undefined)
		gravK = 1;
	if (stiffK == undefined)
		stiffK = .5;
	if (frictionK == undefined)
		frictionK = .1;
	if (isDraggable == undefined)
		isDraggable = true;
	this.isDraggable = isDraggable;
	this.parent = parent;
	parent.ss = this;
	this.gravK = gravK;
	this.stiffK = stiffK;
	this.dampK = 1 - frictionK;
	this.renderFlags = 0;
	this.simFloor = Stage.height;
	this.selectedDot = undefined;
	this.autoReverse = true;
	this.springThickness = 2;
	this.floorDamping = 1;
	this.springColor = 0x000000;
	this.springAlpha = 50;
	this.nbrDots = 0; // only used for internal dot mcs
	this.editMode = 0;
	this.timeStart = getTimer();
	this.timeSnap = 0;
	this.lastReverse = -2;
	this.reset();
	this.setEditMode(0);
}

SpringSim.RENDER_SPRINGS 	= 1;
SpringSim.RENDER_DOTS 		= 2;
SpringSim.RENDER_CURVY 		= 4;
SpringSim.RENDER_POLYS 		= 8;

SpringSim.prototype.setRenderFlags = function(f)
{
	this.renderFlags = f;
	this.parent.clear();
	var showDots = (this.renderFlags & this.RENDER_DOTS) != 0;
	for (var i = 0; i < this.dots.length; ++i) {
		this.dots[i].showDot = showDots;
		this.dots[i].drawDot();
	}
	this.drawSprings();

}

SpringSim.prototype.reset = function()
{
	for (var i = 0; i < this.nbrDots; ++i) {
		var mc = _root["dot"+i];
		mc.removeMovieClip();
	}
	this.nbrDots = 0;
	this.springs = [];
	this.dots = [];
	this.polys = [];
	this.drawSprings();
}

SpringSim.prototype.setGravity = function(gravK)
{
	this.gravK = gravK;
}

SpringSim.prototype.setDrawCurvy = function(s)
{
	if (s)
		this.renderFlags |= SpringSim.RENDER_CURVY;
	else
		this.renderFlags &= ~SpringSim.RENDER_CURVY;
	this.setRenderFlags(this.renderFlags);
}


SpringSim.prototype.setGravity = function(gravK)
{
	this.gravK = gravK;
}

SpringSim.prototype.setStiffness = function(stiffK)
{
	this.stiffK = stiffK;
}

SpringSim.prototype.setFriction = function(frictionK)
{
	this.dampK = 1-frictionK;
}

SpringSim.prototype.setFloorRoughness = function(frictionK)
{
	this.floorDamping = 1-frictionK;
}

SpringSim.prototype.setWave = function(amp,phase,speed)
{
	this.waveAmp = amp;
	this.wavePhase = phase;
	this.waveSpeed = speed;
	this.waveDir = 1;
	this.timeSnap = 0;
	this.timeStart = getTimer();
}


SpringSim.prototype.doReverse = function()
{
	this.timeSnap = this.timeSnap + (getTimer() - this.timeStart)*this.waveDir;
	this.waveDir *= -1;
	this.timeStart = getTimer();
	for (var i = this.dots.length-1; i >= 0; --i)  
	{
		var mc = this.dots[i];
		mc.vx *= -1;
	}
}

SpringSim.prototype.setAutoReverse = function(t)
{
	this.autoReverse = t;
}

SpringSim.prototype.addSpring = function(dot1, dot2, tension)
{
	this.registerDot(dot1);
	this.registerDot(dot2);
	var spr = new Spring(dot1, dot2, tension);
	this.springs.push(spr);
	this.drawSprings();
	return spr;
}

SpringSim.prototype.findSpring = function(dot1, dot2)
{
	for (var i = 0; i < this.springs.length; ++i) {
		var spr = this.springs[i];
		if (spr.dot1 == dot1 && spr.dot2 == dot2)
			return spr;
		if (spr.dot2 == dot1 && spr.dot1 == dot2)
			return spr;
	}
	return undefined;
}

SpringSim.prototype.addPolygon = function(pts, clr, curvy)
{
	var sprs = [0];
	for (var i = 1; i < pts.length; ++i)
		sprs[i] = this.findSpring(pts[i-1], pts[i]);
	var poly = new Polygon(pts, clr, curvy,sprs);
	this.polys.push(poly);
	this.drawSprings();
	return poly;
}



SpringSim.prototype.hasMC = function(mc)
{
	for (var i = this.dots.length-1; i >= 0; --i) 
	{
		if (this.dots[i] == mc)
			return true;
	}
	return false;
}

SpringSim.prototype.registerDot = function(mc)
{
	if (!this.hasMC(mc))
	{
		if (this.isAnchor == undefined)
			this.isAnchor = false;
		this.dots.push(mc);
		mc.drawDot();
		if (this.isDraggable)
		{
			mc.onRollOver = function() { _root.ss.isOverDot = true;  this.hilite = true; this.drawDot(); }
			mc.onRollOut = function() { _root.ss.isOverDot = false;  this.hilite = false; this.drawDot(); }
			mc.onPress = mc.pressDot;
			mc.onReleaseOutside = mc.onRelease = mc.releaseDot;
		}
	}
}

SpringSim.prototype.toggleShowGuide = function(s)
{
	if (s)
		this.renderFlags |= (SpringSim.RENDER_SPRINGS | SpringSim.RENDER_DOTS);
	else
		this.renderFlags &= ~(SpringSim.RENDER_SPRINGS | SpringSim.RENDER_DOTS);
	this.setRenderFlags(this.renderFlags);
}

SpringSim.prototype.drawSprings = function()
{
	if (this.renderFlags == 0)
		return;
	this.parent.clear();
	if (this.renderFlags & SpringSim.RENDER_SPRINGS)
	{
		this.parent.lineStyle(this.springThickness,this.springColor, this.springAlpha);
		if (this.editMode == 1 && this.dragDot != undefined)
		{
			for (var i = this.springs.length-1; i >= 0; --i) 
			{
				var spr = this.springs[i];
				if (spr.dot1 == this.dragDot ||
					spr.dot2 == this.dragDot) 
				{
					spr.dist = spr.restLen = spr.getLength();
				}
			}
		}
		for (var i = this.springs.length-1; i >= 0; --i) 
		{
			var spr = this.springs[i];
			var mc1 = spr.dot1;
			var mc2 = spr.dot2;
			this.parent.moveTo(mc1._x, mc1._y);
			if (!(this.renderFlags & SpringSim.RENDER_CURVY) || mc1.isAnchor || mc2.isAnchor)
				this.parent.lineTo(mc2._x, mc2._y);
			else
			{
				var dx = mc2._x - mc1._x;
				var dy = mc2._y - mc1._y;
				var cx = (mc1._x + mc2._x)/2;
				var cy = (mc1._y + mc2._y)/2;
				var ang = Math.atan2(dy,dx)+Math.PI/2;
				var r = spr.dist - spr.restLen;
				this.parent.curveTo(cx+Math.cos(ang)*r, cy+Math.sin(ang)*r,mc2._x, mc2._y);		
			}
		}
	}
	if (this.renderFlags & SpringSim.RENDER_POLYS)
	{
		var len = this.polys.length;
		for (var i = 0; i < len; ++i) 
		{
			var poly = this.polys[i];
			this.parent.beginFill(poly.clr, 100);
			var mc1 = poly.pts[0];
			var mc2;
			this.parent.moveTo(mc1._x, mc1._y);
			for (var p = 1; p < poly.pts.length; ++p) {
				mc2 = poly.pts[p];
				if (!poly.curvy || poly.sprs[p] == undefined)
					this.parent.lineTo(mc2._x, mc2._y);
				else {
					var spr = poly.sprs[p];
					var dx = mc2._x - mc1._x;
					var dy = mc2._y - mc1._y;
					var cx = (mc1._x + mc2._x)/2;
					var cy = (mc1._y + mc2._y)/2;
					var ang = Math.atan2(dy,dx)+Math.PI/2;
					var r = spr.dist - spr.restLen;
					this.parent.curveTo(cx+Math.cos(ang)*r, cy+Math.sin(ang)*r,mc2._x, mc2._y);		
				}
				mc1 = mc2;
			}
			this.parent.endFill();
		}
	}
}

// DOSIMULATE
// THIS IS WHERE THE FUN PHYSICS ARE: RIGHT HERE IN THIS VERY FUNCTION
//
// Basically we compute the forces affecting each MC, and then
// use them to tweak the motion
//
// Note: Unlike many physics code examples, we completely ignore
// time, assuming that frame-rate is relatively constant (it is close enough).
//
// For our purposes this is just fine -it makes the code run faster
// and makes it more readable.
SpringSim.prototype.doSimulate = function()
{
	var elapsed = this.timeSnap + (getTimer() - this.timeStart)*this.waveDir;
	var waveIdx = elapsed * 0.1 * Math.PI * this.waveSpeed + this.wavePhase;
	// trace("sim " + this.dots.length + ", " + this.springs.length);
	// zero out forces & add gravity
	for (var i = this.dots.length-1; i >= 0; --i) 
	{
		var mc = this.dots[i];
		mc.fx = 0;
		mc.fy = 0;
		if (!mc.isAnchor && !mc.isDragging)
			mc.fy = this.gravK;
	}

	// figure out the amount and direction of the force the springs are applying to the particles
	for (var i = this.springs.length-1; i >= 0; --i) 
	{
		var spr = this.springs[i];
		var mc1 = spr.dot1;
		var mc2 = spr.dot2;

		// distance between spring endpoints
		var dist = spr.getLength();
		spr.dist = dist;

		var f1 = this.stiffK * (dist - spr.restLen);
		if (spr.isMuscle) 
			f1 += .5*spr.restLen*spr.amp*this.waveAmp*Math.sin(waveIdx+spr.phase);
		// spring force is proportional to difference with resting length
				// HERE BE HOOKE'S LAW
				// Robert Hooke - A great man indeed :)  Also drew a mean flea!
				// If only he had flash...

		// and is applied in direction of spring
		var spr_fx = f1 * (mc2._x - mc1._x)/dist;
		var spr_fy = f1 * (mc2._y - mc1._y)/dist;

		// springs only affect 'loose' MCs
		if (!mc1.isAnchor && !mc1.isDragging) 
		{
			mc1.fx += spr_fx;
			mc1.fy += spr_fy;
		}
		// spring has opposite affect on each connected MC
		if (!mc2.isAnchor && !mc2.isDragging) 
		{
			mc2.fx -= spr_fx;
			mc2.fy -= spr_fy
		}
	}

	// For each MC
	//
	var minx = Stage.width;
	var maxx = 0;
	for (var i = this.dots.length-1; i >= 0; --i)  
	{
		var mc = this.dots[i];
		if (mc.isDragging)
			continue;

		// Applying damping (friction, viscosity - whatever) - this is what
		// slows us down
		mc.vx *= this.dampK;
		mc.vy *= this.dampK;

		// Apply spring forces and gravity to the MC's current velocity
		mc.vx += mc.fx;
		mc.vy += mc.fy;

		if (mc._y+mc.vy > this.simFloor) {
			mc.vy -= mc._y+mc.vy - this.simFloor;
			mc.vy *= this.floorDamping;
			mc.vx *= this.floorDamping;
		}
		// Use the velocity to adjust our position
		//
		mc._x += mc.vx;	
		mc._y += mc.vy;	
		maxx = Math.max(maxx, mc._x);
		minx = Math.min(minx, mc._x);
	}
	if ((maxx > Stage.width && this.lastReverse != 1) || (minx < 10 && this.lastReverse != 0)) {
		this.lastReverse = (maxx > Stage.width);
		this.doReverse();
	}
	// Redraw the springs (the guide), if necessary
	this.drawSprings();
}

// Editing functions - not used for basic spring sim stuff
//

// use this if you want SpringSim to provide the movieclips for the dots
// you may also provide your own MCs
SpringSim.prototype.addDotMC = function(x,y,isFixed,layer)
{
	var i = this.getDotNbr();
	var mc = this.parent.createEmptyMovieClip("dot" + i, layer);
	mc._x = x;
	mc._y = y;
	mc.dotIdx = i;
	mc.setDotProps(isFixed,true);
	mc.showDot();
	this.registerDot(mc);
	return mc;
}

SpringSim.prototype.addAMouseDot = function(layer)
{
	if (this.parent._ymouse < 300 && this.parent._ymouse > 0 &&
		this.parent._xmouse < SW && this.parent._xmouse > 0)
	{
		var mc = this.addDotMC(this.parent._xmouse, this.parent._ymouse, false, layer);
		if (this.selectedDot != undefined)
			this.addSpring(this.selectedDot, mc);
		this.doSelectDot(mc);
		mc.creationTime = getTimer();
		return mc;
	}
	return undefined;
}


SpringSim.prototype.getDotNbr = function()
{
	for (var i = 0; i < this.nbrDots; ++i) {
		if (_root["dot"+i] == undefined)
			return i;
	}
	return this.nbrDots++;
}

SpringSim.prototype.showDots = function(s) // affects built-ins only
{
	for (var i = 0; i < this.nbrDots; ++i) {
		var mc = _root["dot"+i];
		if (mc != undefined)
			mc.setDotProps(mc.isAnchor, s);
	}
}


SpringSim.prototype.doSelectDot = function(dot)
{
	if (this.selectedDot != undefined) {
		this.selectedDot.isSelected = false;
		this.selectedDot.drawDot();
	}
	if (dot != undefined) {
		this.selectedDot = dot;
		dot.isSelected = true;
		dot.drawDot();
	}
	else
		this.selectedDot = undefined;
}

SpringSim.prototype.deleteDot = function(dot)
{
	for (var i = this.springs.length-1; i >= 0; --i) {
		if (this.springs[i].dot1 == dot || this.springs[i].dot2 == dot)
			this.springs.splice(i,1);
	}
	dot.removeMovieClip();
	this.drawSprings();
	this.doSelectDot(undefined);
}

SpringSim.prototype.setEditMode = function(f)
{
	// trace("edit mode = " + f);
	this.editMode = f;
	switch (this.editMode) {
	case 0:	this.doSelectDot(undefined);
			this.parent.onEnterFrame = function() { this.ss.doSimulate(); }
			break;
	case 1:	_root.onEnterFrame = undefined;
			break;
	case 2:	_root.onEnterFrame = undefined;
			break;
	}
}



// Movieclip stuff...
//

// This is what happens when we start dragging an anchor point
//
MovieClip.prototype.pressDot = function()
{
	if (_root.ss.editMode == 0) {	// normal simulation
		this.isDragging = true;
		_root.ss.dragDot = this;
		_root.ss.isOverDot = true;
		this.startDrag(false);
	}
	else {							// edit mode / delete mode
		this.ox = this._x;
		this.oy = this._y;
		this.lastSelect = _root.ss.selectedDot;
		_root.ss.doSelectDot(undefined);
		this.isDragging = true;
		_root.ss.dragDot = this;
		_root.ss.isOverDot = true;
		this.startDrag(false);
		_root.onEnterFrame = function() { this.ss.drawSprings(); }
	}
}

// This is what happens when we start release an anchor point
//
MovieClip.prototype.releaseDot = function()
{
	switch (_root.ss.editMode) {
	case 0:						// normal simulation
		this.stopDrag();
		this.isDragging = false;
		_root.ss.dragDot = undefined;
		break;
	case 1:						// edit mode
		_root.onEnterFrame = undefined;
		this.stopDrag();
		var wasDragging = this.isDragging;
		this.isDragging = false;
		_root.ss.dragDot = undefined;
		var wasDragged = Math.abs(this.ox - this._x) >= 3 || Math.abs(this.oy - this._y) >= 3;
		var newlyCreated = getTimer() - this.creationTime < 1000;
		if (!wasDragged && this.lastSelect != undefined && this.lastSelect != this && !this.hasSpringTo(this.lastSelect))
			_root.ss.addSpring(this.lastSelect, this);
			
		if (this.lastSelect != this || wasDragged || newlyCreated) {
			_root.ss.doSelectDot(this);
		}
		break;
	case 2:						// delete mode
		_root.ss.deleteDot(this);
		break;
	}
}

MovieClip.prototype.drawCircle = function(x,y,radius)
{
	var r = radius;
	var n= 8;
	
	var theta = 45*kDegToRad;
	var cr = radius/Math.cos(theta/2);
	var angle = bA*kDegToRad;
	var cangle = angle-theta/2;
	
	this.moveTo(x+r, y);
	for (var i=0;i < n;i++) 
	{
		angle += theta;
		cangle += theta;
		var endX = r*Math.cos (angle);
		var endY = r*Math.sin (angle);
		var cX = cr*Math.cos (cangle);
		var cY = cr*Math.sin (cangle);
		this.curveTo(x+cX,y+cY, x+endX,y+endY);
	}
}

MovieClip.prototype.setDotProps = function(isAnchor,renderDot)
{
	this.isAnchor = isAnchor;
	this.renderDot = renderDot;
	this.vx = this.vy = 0;
	this.drawDot();
	_root.ss.drawSprings();
}

SpringSim.dotRad = 5;

MovieClip.prototype.drawDot = function()
{
	this.clear();
	this.beginFill((this.isAnchor? 0x770000 : 0x000077), this.renderDot? 100 : 0);
	this.drawCircle(0,0, this.hilite? SpringSim.dotRad*1.5 : SpringSim.dotRad);
	this.endFill();
	if (this.isSelected) 
	{
		this.lineStyle(1, 0xFF7700,100);
		this.drawCircle(0,0, SpringSim.dotRad*2);
	}
}

MovieClip.prototype.hasSpringTo = function(mc)
{
	for (var i = _root.ss.springs.length-1; i >= 0; --i)
	{
		var spr = _root.ss.springs[i];
		if ((spr.dot1 == mc && spr.dot2 == this) ||
			(spr.dot1 == this && spr.dot2 == mc)) 
		{
			// trace("spring already exists");
			return true;
		}
	}
	return false;
}
